import pandas as pd

def remove_out_of_bounds_bins(df, chromosome_size):
    # type: (pd.DataFrame, int) -> pd.DataFrame
    """Remove all reads that were shifted outside of the genome endpoints."""

    # The dataframe is empty and contains no bins out of bounds
    if "Bin" not in df:
        return df

    df = df.drop(df[df.Bin > chromosome_size].index)

    return df.drop(df[df.Bin < 0].index)


def remove_bins_with_ends_out_of_bounds(df, chromosome_size,
                                        window_size):
    # type: (pd.DataFrame, int, int) -> pd.DataFrame
    """Remove all reads that were shifted outside of the genome endpoints."""

    # The dataframe is empty and contains no bins out of bounds

    # print(df.head(2))
    # print(chromosome_size)
    # print(window_size)
    out_of_bounds = df[df.index.get_level_values("Bin") + window_size >
                       chromosome_size].index
    # print(len(out_of_bounds))
    df = df.drop(out_of_bounds)

    return df

    # dfms = Parallel(n_jobs=args.number_cores)(
    #     delayed(_create_matrixes)(chromosome, chip, input, islands)
    #     for chromosome in all_chromosomes)






from joblib import Parallel, delayed
from collections import defaultdict
import pandas as pd
import numpy as np

import pkg_resources, os
from natsort import natsorted

from io import StringIO

from typing import DefaultDict, Dict, Iterable
# from helper.functions

import logging
from rpy2.robjects import r, pandas2ri
pandas2ri.activate()
from rpy2.robjects.robject import RObject

from rpy2.robjects.packages import importr

importr("S4Vectors")
bioc = importr("GenomicRanges")



def files_to_chromosome_coverage(all_files, nb_cpu):
    # type: (Iterable[str], int) -> DefaultDict[str,Dict[str, RObject]]

    df_to_coverage = r("function(x) coverage(GRanges(x$Chromosome, IRanges(x$Start, x$End)))")

    logging.info("Finding nucleotide coverage of files.")
    coverages = defaultdict(dict) # type: DefaultDict[str,Dict[str, RObject]]
    for f in all_files:
        df = pd.read_table(f, usecols=[0, 1, 2], header=None, names="Chromosome Start End".split())
        cv = df_to_coverage(df)

        chromosomes = r["names"](cv)
        for chromosome in chromosomes:
            coverages[f][chromosome] = r("function(x, idx) x[[idx]]")(cv, chromosome)

    max_per_chromosome_coverage = r("function(x) sum(runLength(x))")
    remove_duplicate_list_entries = r("function(x) x[unique(names(x))]")

    maxlengths = defaultdict(int) # type: DefaultDict[str, int]
    for f, data in coverages.items():
        for chromosome, rle in data.items():
            current_len = max_per_chromosome_coverage(rle)[0]
            maxlengths[chromosome] = max(maxlengths[chromosome], current_len)

    extend_rle = r('function(cvg, maxlen) c(cvg,Rle(0,maxlen-length(cvg)))')
    extended_rles = defaultdict(dict) # type: DefaultDict[str,Dict[str, RObject]]
    for f, d in coverages.items():
        for chromosome, cv in d.items():
            maxlength = maxlengths[chromosome]
            cv = extend_rle(cv, maxlength)
            extended_rles[f][chromosome] = cv

    return extended_rles



import pytest

from joblib import Parallel, delayed
from collections import defaultdict
import pandas as pd
import numpy as np
from typing import Dict, Iterable

import pkg_resources, os
from natsort import natsorted

from io import StringIO

# from helper.functions

import logging
from rpy2.robjects import r, pandas2ri
pandas2ri.activate()
from rpy2.robjects.robject import RObject

from rpy2.robjects.packages import importr
importr("S4Vectors")
bioc = importr("GenomicRanges")

length_of_rle = r("function (x) sum(runLength(x))")


from epic.scripts.overlaps.files_to_chromosome_coverage import (files_to_chromosome_coverage)

__author__ = "Endre Bakken Stovner https://github.com/endrebak/"
__license__ = "MIT"


def nucleotide_overlaps_per_file(all_files, nb_cpu):
    # type: (Iterable[str], int) -> pd.DataFrame

    rles = files_to_chromosome_coverage(all_files, nb_cpu)

    nucleotide_overlaps = Parallel(n_jobs=nb_cpu)(delayed(_nucleotide_overlaps_per_file)(
        f, rles) for f in rles)

    print(nucleotide_overlaps)
    return pd.concat(nucleotide_overlaps).sort_values(["Main", "Other"]).reset_index(drop=True)


def _nucleotide_overlaps_per_file(bed_file, extended_rles):
    # type: (str, Dict[str,Dict[str, RObject]]) -> pd.DataFrame

    base_bed = bed_file.split("/")[-1].split(".")[0]
    logging.info("Finding the number of nucleotides in " + base_bed + " overlapping other files.")

    _find_overlaps = r("""
    function(s, o) {
    runValue(s) = as.logical(runValue(s))
    runValue(o) = as.logical(runValue(o))
    sum(s & o)
    }
    """)

    _find_total = r(" function(s) {runValue(s) = as.logical(runValue(s)); sum(s) }")

    cvs = extended_rles[bed_file]

    rowdicts = []
    for f in extended_rles:

        base_bed_other = f.split("/")[-1].split(".")[0]
        # print("base bed other", base_bed_other)
        cvos = extended_rles[f]

        overlapping_chromosomes = set(cvs.keys()).intersection(cvos.keys())
        overlaps, total = 0, 0
        for c in overlapping_chromosomes:

            cv, cvo = cvs[c], cvos[c]

            fov = _find_overlaps(cv, cvo)[0]
            tot = _find_total(cv)[0]

            overlaps += fov
            total += tot

        ratio = overlaps/total
        rowdict = {"Chromosome": c, "Main": base_bed, "Other": base_bed_other, "Overlaps": ratio}
        rowdicts.append(rowdict)

    return pd.DataFrame.from_dict(rowdicts)["Chromosome Main Other Overlaps".split()]



import logging
from natsort import natsorted

from joblib import Parallel, delayed
import pandas as pd

from typing import Iterable, Sequence, Tuple
try:
    from functools import lru_cache # type: ignore
except ImportError:
    from functools32 import lru_cache # type: ignore


def _merge_chip_and_input(chip_df, input_df):
    # type: (pd.DataFrame, pd.DataFrame) -> pd.DataFrame

    chip_df = chip_df.set_index("Chromosome Bin".split())
    input_df = input_df.set_index("Chromosome Bin".split())

    chip_df_nb_bins = len(chip_df)

    merged_df = chip_df.join(input_df,
                             how="left",
                             lsuffix=" ChIP",
                             rsuffix=" Input").reset_index()
    merged_df = merged_df[["Chromosome", "Bin", "Count ChIP", "Count Input"]]
    merged_df.columns = ["Chromosome", "Bin", "ChIP", "Input"]

    merged_df = merged_df.fillna(0)

    if not len(merged_df) == chip_df_nb_bins:
        assertion_message = [
            "Wrong number of rows after merging ChIP/Input.",
            "ChIP bins: " + str(chip_df_nb_bins),
            "Input bins: " + str(len(input_df)), "Head of ChIP df: ",
            chip_df.head().to_csv(sep=" "), "Head of Input df: ",
            input_df.head().to_csv(sep=" "), "Tail of ChIP df: ",
            chip_df.tail().to_csv(sep=" "), "Tail of Input df: ",
            input_df.tail().to_csv(sep=" "), "Number of bins in merged df: ",
            str(len(merged_df)), "Head of merged df: ",
            merged_df.head().to_csv(sep=" "), "Tail of merged df: ",
            merged_df.tail().to_csv(sep=" ")
        ]
        assertion_message = "\n".join(assertion_message) # type: ignore
        assert len(merged_df) == chip_df_nb_bins, assertion_message

    return merged_df


def merge_chip_and_input(chip_dfs, input_dfs, nb_cpu):
    # type: (Iterable[pd.DataFrame], Iterable[pd.DataFrame], int) -> Sequence[pd.DataFrame]

    # should be same length, since missing chromos get empty df
    # assert len(chip_dfs) == len(input_dfs)
    # if len(chip_dfs) != len(input_dfs):
    #     logging.info("Different number of chromosomes in ChIP and Input.")
    #     logging.info("Chromosomes in ChIP:" chip)

    logging.info("Merging ChIP and Input data.")
    merged_chromosome_dfs = Parallel(n_jobs=nb_cpu)(
        delayed(_merge_chip_and_input)(chip_df, input_df)
        for chip_df, input_df in zip(chip_dfs, input_dfs))
    return merged_chromosome_dfs


def get_total_number_of_reads(dfs):
    # type: (Iterable[pd.DataFrame]) -> int
    return sum([df.Count.sum() for df in dfs])


def ensure_same_chromosomes_in_list(sample1_dfs, sample2_dfs):
    # type: (List[pd.DataFrame], List[pd.DataFrame]) -> Tuple[Dict[str, pd.DataFrame], Dict[str, pd.DataFrame]]

    d1 = create_chromsome_df_map(sample1_dfs)
    d2 = create_chromsome_df_map(sample2_dfs)

    d1, d2 = fill_missing_chromosomes(d1, d2)

    assert set(d1.keys()) == set(d2.keys())

    return d1, d2


def create_chromsome_df_map(dfs):
    # type: (Iterable[pd.DataFrame]) -> Dict[str, pd.DataFrame]

    sample_dict = {}
    for df in dfs:

        if df.empty:
            continue

        chromosome = df.head(1).Chromosome.values[0]
        sample_dict[chromosome] = df

    return sample_dict


def fill_missing_chromosomes(d1, d2):
    # type: (pd.DataFrame, pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]

    all_chromosomomes = set(d1.keys()).union(d2.keys())

    d1_missing = set(d2.keys()).difference(d1.keys())
    d2_missing = set(d1.keys()).difference(d2.keys())

    for chromosome in d1_missing:
        d1[chromosome] = pd.DataFrame(columns=["Chromosome", "Bin"])

    for chromosome in d2_missing:
        d2[chromosome] = pd.DataFrame(columns=["Chromosome", "Bin"])

    return d1, d2


def merge_same_files(sample1_dfs, sample2_dfs, nb_cpu):
    # type: (List[pd.DataFrame], List[pd.DataFrame], int) -> List[pd.DataFrame]

    # if one list is missing a chromosome, we might pair up the wrong dataframes
    # therefore creating dicts beforehand to ensure they are paired up properly
    d1, d2 = ensure_same_chromosomes_in_list(sample1_dfs,
                                             sample2_dfs)

    assert len(d1) == len(d2)

    logging.info("Merging same class data.")
    merged_chromosome_dfs = Parallel(n_jobs=nb_cpu)(delayed(_merge_same_files)(
        d1[chromosome],
        d2[chromosome]) for chromosome in d1.keys())

    return merged_chromosome_dfs


def _merge_same_files(sample1_df, sample2_df):
    # type: (pd.DataFrame, pd.DataFrame) -> pd.DataFrame

    merged_df = sample1_df.merge(sample2_df,
                                 how="outer",
                                 on=["Chromosome", "Bin"])
    # merged_df = merged_df[~merged_df.index.duplicated(keep='first')]

    return merged_df.fillna(0)



import pytest

from joblib import Parallel, delayed
from collections import defaultdict
import pandas as pd
import numpy as np
from typing import DefaultDict, Dict, Iterable

import pkg_resources, os
from natsort import natsorted

from io import StringIO

# from helper.functions

import logging
from rpy2.robjects import r, pandas2ri
pandas2ri.activate()
from rpy2.robjects.robject import RObject

from rpy2.robjects.packages import importr
importr("S4Vectors")
bioc = importr("GenomicRanges")

from epic.scripts.overlaps.files_to_chromosome_coverage import files_to_chromosome_coverage

__author__ = "Endre Bakken Stovner https://github.com/endrebak/"
__license__ = "MIT"

def overlap_matrix_nucleotides(all_files, nb_cpu):
    # type: (Iterable[str], int) -> pd.DataFrame
    rles = files_to_chromosome_coverage(all_files, nb_cpu)

    nucleotide_overlaps = Parallel(n_jobs=nb_cpu)(delayed(_overlap_matrix_nucleotides)(
        f, rles) for f in rles)

    df = pd.concat(nucleotide_overlaps)

    return df.reset_index(drop=True)


def _overlap_matrix_nucleotides(bed_file, extended_rles):
    # type: (str, Dict[str,Dict[str, RObject]]) -> pd.DataFrame

    overlaps = _create_overlap_matrix_nucleotides(bed_file, extended_rles)
    return _counts_runlengths(bed_file, overlaps)





def _create_overlap_matrix_nucleotides(bed_file, coverages):
    # type: (str, Dict[str,Dict[str, RObject]]) -> Dict[str, RObject]

    base_bed_other = bed_file.split("/")[-1]
    logging.info("Processing {} at nucleotide level".format(base_bed_other))

    cvs = coverages[bed_file]

    files = [f for f in coverages if f != bed_file]
    compute_overlap = r("function(x, o) x + (x & o)")
    for f in files:
        # print(f)

        other_cvs = coverages[f]
        overlapping_chromosomes = set(cvs.keys()).intersection(other_cvs.keys())

        for chromosome in overlapping_chromosomes:
            chr_cov = cvs[chromosome]
            other_cov = other_cvs[chromosome]

            overlap = compute_overlap(chr_cov, other_cov)
            cvs[chromosome] = overlap

    return cvs


def _counts_runlengths(bed_file, cvs):
    # type: (str, Dict[str, RObject]) -> pd.DataFrame

    base_bed = bed_file.split("/")[-1].split(".")[0]

    overlaps = defaultdict(int) # type: DefaultDict[str, int]
    get_runlength = r("function(x, v) runLength(x[x == v])")
    for chromosome, overlap in cvs.items():
        run_values = set(r["runValue"](overlap))
        for run_value in run_values:
            run_value = int(run_value)
            runlength = get_runlength(overlap, run_value)[0]
            overlaps[run_value] += runlength

    rowdicts = []
    for run_value, run_length in overlaps.items():
        rowdict = {"Main": base_bed, "Other": run_value, "Overlaps": run_length}
        rowdicts.append(rowdict)

    return pd.DataFrame.from_dict(rowdicts)



import logging
from scipy.stats import poisson
from numpy import log

from epic.config.constants import BIN_SIZE, E_VALUE_THRESHOLD
from epic.statistics.generate_cumulative_distribution import generate_cumulative_dist
from epic.statistics.add_to_island_expectations import add_to_island_expectations_dict


def compute_score_threshold(average_window_readcount,
                            island_enriched_threshold,
                            gap_contribution, boundary_contribution,
                            genome_length_in_bins):
    # type: (float, int, float, float, float) -> float
    """
    What does island_expectations do?
    """

    required_p_value = poisson.pmf(island_enriched_threshold,
                                   average_window_readcount)

    prob = boundary_contribution * required_p_value

    score = -log(required_p_value)

    current_scaled_score = int(round(score / BIN_SIZE))

    island_expectations_d = {}  # type: Dict[int, float]
    island_expectations_d[current_scaled_score] = prob * genome_length_in_bins
    island_expectations_d[
        0] = boundary_contribution * genome_length_in_bins / gap_contribution

    current_max_scaled_score = current_scaled_score

    interval = int(1 / BIN_SIZE)
    partial_cumu = 0.0
    logging.info("Finding the score required to consider an island enriched.")
    while (partial_cumu > E_VALUE_THRESHOLD or partial_cumu < 1e-100):

        current_scaled_score += interval
        current_max_scaled_score = current_scaled_score - interval
        # logging.debug(island_expectations_d)

        if current_scaled_score > current_max_scaled_score:

            # logging.debug(island_expectations_d)
            island_expectations_d = add_to_island_expectations_dict(
                average_window_readcount, current_max_scaled_score,
                island_enriched_threshold, island_expectations_d,
                gap_contribution)
            partial_cumu = 0.0001
            current_max_scaled_score += 1000

            if max(island_expectations_d) > interval:
                partial_cumu = sum(
                    [val
                     for idx, val in island_expectations_d.items()
                     if idx > current_max_scaled_score - interval])
            else:
                partial_cumu = sum(island_expectations_d.values())

    logging.debug("Computing cumulative distribution.")
    score_threshold = generate_cumulative_dist(island_expectations_d,
                                               current_max_scaled_score + 1)
    logging.info("Enriched score threshold for islands: " + str(
        score_threshold))
    return score_threshold



from io import StringIO
from subprocess import check_output
from os.path import basename
import logging
from typing import Iterable

import pandas as pd
from joblib import Parallel, delayed


def overlap_matrix_region_counts(all_files, nb_cpu):
    # type: (Iterable[str], int) -> pd.DataFrame

    # bargraph
    regions_matrixes = Parallel(n_jobs=nb_cpu)(
        delayed(_create_overlap_matrix_regions)(bed_file, all_files)
        for bed_file in all_files)

    overlaps = []
    for df in regions_matrixes:
        overlaps.append(_compute_region_overlap(df))

    regions_df = pd.concat(overlaps).reset_index(drop=True)

    return regions_df


def _create_overlap_matrix_regions(bed_file, all_files):
    # type: (str, Iterable[str]) -> pd.DataFrame

    all_files_str = " ".join(all_files)

    base_bed = basename(bed_file).split(".")[0]
    logging.info("Processing {}".format(base_bed))
    command = "bedtools intersect -wo -a {} -b {} -filenames".format(bed_file, all_files_str)
    output = check_output(command, shell=True).decode()

    df = pd.read_table(StringIO(output), usecols=[0, 1, 2, 6], header=None, names="Chromosome Start End Other".split())
    df.Other = df.Other.apply(basename).str.split(".", expand=True).ix[:, 0]

    df.insert(0, "Main", base_bed)

    return df


def _compute_region_overlap(df):
    # type: (pd.DataFrame) -> pd.DataFrame

    main_file = df.Main.ix[0,0]

    nb_overlap = df.drop_duplicates("Chromosome Start End Other".split()).groupby("Chromosome Start End".split()).size().value_counts().to_frame().reset_index()

    nb_overlap.columns = "Other Overlaps".split()

    nb_overlap.insert(0, "Main", main_file)

    return nb_overlap


def overlap_matrix_regions(all_files, nb_cpu):
    # type: (Iterable[str], int) -> pd.DataFrame

    #heatmap

    regions_matrixes = Parallel(n_jobs=nb_cpu)(
        delayed(_create_overlap_matrix_regions)(bed_file, all_files)
        for bed_file in all_files)

    df = pd.concat(regions_matrixes).reset_index(drop=True).drop_duplicates()
    df.to_csv("useme.csv", sep=" ")

    nb_main = df.groupby("Main Chromosome Start End".split()).size().to_frame().reset_index().groupby("Main").size()
    nb_other = df.groupby("Main Other".split()).size()

    result = (nb_other / nb_main).reset_index()
    result.columns = "Main Other Overlaps".split()

    return result



"""Run whole epic pipeline."""

__author__ = "Endre Bakken Stovner https://github.com/endrebak/"
__license__ = "MIT"

from os.path import dirname, join, basename
from sys import stdout
from itertools import chain
from collections import OrderedDict
from subprocess import call
import logging

from argparse import Namespace
import pandas as pd
from numpy import log2

from typing import Iterable
from natsort import natsorted
from joblib import Parallel, delayed

from epic.windows.count.count_reads_in_windows import (
    count_reads_in_windows, count_reads_in_windows_paired_end)
from epic.statistics.compute_background_probabilites import compute_background_probabilities
from epic.statistics.count_to_pvalue import count_to_pvalue
from epic.statistics.fdr import compute_fdr
from epic.utils.helper_functions import merge_chip_and_input, get_total_number_of_reads, merge_same_files
from epic.windows.cluster.find_islands import find_islands
from epic.matrixes.matrixes import write_matrix_files


def run_epic(args):
    # type: (Namespace) -> pd.DataFrame

    chip_windows = multiple_files_count_reads_in_windows(args.treatment, args)
    input_windows = multiple_files_count_reads_in_windows(args.control, args)

    chip_merged = _merge_files(chip_windows.values(), args.number_cores)
    input_merged = _merge_files(input_windows.values(), args.number_cores)

    chip_merged_sum = sum_columns(chip_merged)
    input_merged_sum = sum_columns(input_merged)

    nb_chip_reads = get_total_number_of_reads(chip_merged_sum)
    nb_input_reads = get_total_number_of_reads(input_merged_sum)

    merged_dfs = merge_chip_and_input(chip_merged_sum, input_merged_sum,
                                      args.number_cores)

    score_threshold, island_enriched_threshold, average_window_readcount = \
        compute_background_probabilities(nb_chip_reads, args)

    dfs = []                    # type: Iterable[pd.DataFrame]
    dfs = count_to_pvalue(merged_dfs, island_enriched_threshold,
                          average_window_readcount, args.number_cores)

    dfs = find_islands(dfs, score_threshold, args)

    logging.info("Done finding islands.")
    logging.info("Concating dfs.")
    df = pd.concat([df for df in dfs if not df.empty])
    logging.info("Labeling island bins.")

    logging.info("Computing FDR.")
    df = compute_fdr(df, nb_chip_reads, nb_input_reads, args)

    if (args.store_matrix or args.bigwig or args.sum_bigwig):
        write_matrix_files(chip_merged, input_merged, df, args)

    # Just in case some ints got promoted to float somewhere
    df[["Start", "End", "ChIP", "Input"]] = df[["Start", "End", "ChIP", "Input"
                                                ]].astype(int)
    df.to_csv(stdout, index=False, sep=" ", na_rep="NA")

    if args.bed:
        df_to_bed(df).to_csv(args.bed, header=False, index=False, sep="\t")

    return df.reset_index(
    )  # only returns a value to simplify integration tests


def df_to_bed(df):
    # type: (pd.DataFrame) -> pd.DataFrame

    # '''Chromosome Start End ChIP Input Score Fold_change P FDR
    # chr5 53000 55399 121 13 77.6075622841774 13.655736573980159 6.040968494897508e-92 1.9241805908359603e-91\''

    regions = df["Chromosome Start End FDR".split()].copy()
    regions.insert(4, "Score", log2(df.ChIP/df.Input) * 100)
    regions.loc[regions.Score > 1000, "Score"] = 1000
    regions.loc[regions.Score < 0, "Score"] = 0
    regions.insert(5, "Strand", ".")

    return regions




def sum_columns(dfs):
    # type: (Iterable[pd.DataFrame]) -> List[pd.DataFrame]

    new_dfs = []
    for df in dfs:
        s = df.set_index("Chromosome Bin".split())
        s = s.sum(axis=1)
        s.name = "Count"
        df = s.reset_index()
        new_dfs.append(df)

    return new_dfs


def multiple_files_count_reads_in_windows(bed_files, args):
    # type: (Iterable[str], Namespace) -> OrderedDict[str, List[pd.DataFrame]]
    """Use count_reads on multiple files and store result in dict.

    Untested since does the same thing as count reads."""

    bed_windows = OrderedDict() # type: OrderedDict[str, List[pd.DataFrame]]
    for bed_file in bed_files:
        logging.info("Binning " + bed_file)
        if args.paired_end:
            chromosome_dfs = count_reads_in_windows_paired_end(bed_file, args)
        else:
            chromosome_dfs = count_reads_in_windows(bed_file, args)
        bed_windows[bed_file] = chromosome_dfs

    return bed_windows


def _merge_files(windows, nb_cpu):
    # type: (Iterable[pd.DataFrame], int) -> pd.DataFrame
    """Merge lists of chromosome bin df chromosome-wise.

    windows is an OrderedDict where the keys are files, the values are lists of
    dfs, one per chromosome.

    Returns a list of dataframes, one per chromosome, with the collective count
    per bin for all files.

    TODO: is it faster to merge all in one command?
    """

    # windows is a list of chromosome dfs per file
    windows = iter(windows)  # can iterate over because it is odict_values
    merged = next(windows)

    # if there is only one file, the merging is skipped since the windows is used up
    for chromosome_dfs in windows:
        # merge_same_files merges the chromosome files in parallel
        merged = merge_same_files(merged, chromosome_dfs, nb_cpu)

    return merged



from math import log, factorial, pi, exp
from epic.utils.helper_functions import lru_cache


@lru_cache()
def _factln(num):
    # type: (int) -> float
    """
    Computes logfactorial regularly for tractable numbers, uses Ramanujans approximation otherwise.
    """

    if num < 20:
        log_factorial = log(factorial(num))
    else:
        log_factorial = num * log(num) - num + log(num * (1 + 4 * num * (
            1 + 2 * num))) / 6.0 + log(pi) / 2

    return log_factorial


@lru_cache()
def _poisson(i, average):
    # type: (int, float) -> float
    """
    """
    exponent = -average + i * log(average) - _factln(i)
    return exp(exponent)
